/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         Connection.java
**
** Description:  Contains the Connection class that is internal to the KRIL.
**
*/
package kqml;
import java.net.Socket;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.io.EOFException;
import java.util.Date;

/**
 * A class internal to the KRIL.
 *
 * Objects of type "Connection" are associated with KQML client proxies
 * that are executing on the WEB server from whence the executing applet
 * was loaded.  This class is used to perform the I/O to a socket, including
 * buffering of input messages.
 *
 * @author Peter Stevens
 * @version $Revision: 1.7 $
 * @see KqmlAgent
 * @see Message
 */
class Connection
{
    private static final boolean DEBUG = false;

    private KqmlAgent    agent            = null;
    private String       agentName        = null;
    private boolean      connectionIsOpen = false;
    private String       hostName         = null;
    private InputStream  in;
    private String       localHost;
    private OutputStream out;
    private int          port             = 0;
    private int          readCalls        = 0;
    private String       remoteAgent      = null;
    private Socket       sock;
    private int          startIndex       = 0;
    private int          totalBytes       = 1;

    // Constructors

    //
    // Constructor to open a connection following an ANS lookup
    //
    protected Connection(AgentAddress addr, KqmlAgent agent)
                                       throws InvalidKQMLOpenException
    {
        String          hostname;
        int             port;

        hostname = addr.getHostAddress();
        try
        {
            port = Integer.parseInt(addr.getHostPort());
            sock = new Socket(hostname, port);
        }
        catch(java.io.IOException e)
        {
            throw new InvalidKQMLOpenException(
                        " cannot open socket to host '" + hostname + ":" +
                        addr.getHostPort() + "' (" + e.getMessage() + ")");
        }
        catch(NumberFormatException e)
        {
            throw new InvalidKQMLOpenException(
                        " cannot open socket port number not numeric'" +
                        addr.getHostPort() + "' (" + e.getMessage() + ")");
        }

        try
        {
            in = sock.getInputStream();
        }

        catch(java.io.IOException e)
        {
            throw new InvalidKQMLOpenException(
                        " cannot open socket input stream '" + hostname +
                        "' (" + e.getMessage() + ")");
        }

        try
        {
            out = sock.getOutputStream();
        }
        catch(java.io.IOException e)
        {
            throw new InvalidKQMLOpenException(
                         " cannot open socket output stream '" + hostname +
                         " (" + e.getMessage() + ")");
        }
        remoteAgent      = addr.getName();
        agentName        = agent.getAgentName();
        hostName         = hostname;
        this.port        = port;
        this.agent       = agent;
        connectionIsOpen = true;
    }

    //
    // Constructor to open a connection the ANS
    //
    protected Connection(String hostName, int port, String remoteAgent,
                            KqmlAgent agent) throws InvalidKQMLOpenException
    {
        try
        {
            sock = new Socket(hostName, port);
        }
        catch(java.io.IOException e)
        {
            throw new InvalidKQMLOpenException(
                            " cannot open socket to host '" + hostName +
                                      "' (" + e.getMessage() + ")");
        }

        try
        {
            in = sock.getInputStream();
        }
        catch(java.io.IOException e)
        {
            throw new InvalidKQMLOpenException(
                            " cannot open socket input stream '" + hostName +
                                      "' (" + e.getMessage() + ")");
        }

        try
        {
            out = sock.getOutputStream();
        }
        catch(java.io.IOException e)
        {
            throw new InvalidKQMLOpenException(
                           " cannot open socket output stream '" + hostName +
                                       " (" + e.getMessage() + ")");
        }

        this.agent       = agent;
        this.remoteAgent = remoteAgent;
        agentName        = agent.getAgentName();
        this.hostName    = hostName;
        this.port        = port;
        connectionIsOpen = true;
    }

    //
    // Constructor for a connection resulting from an "accept"
    //
    protected Connection(Socket sock, KqmlAgent agent)
                                          throws InvalidKQMLOpenException
    {
        try
        {
            in = sock.getInputStream();
        }

        catch(java.io.IOException e)
        {
            throw new InvalidKQMLOpenException(
                           " cannot open socket input stream '" +
                           "' (" + e.getMessage() + ")");
        }

        try
        {
            out = sock.getOutputStream();
        }
        catch(java.io.IOException e)
        {
            throw new InvalidKQMLOpenException(
                           " cannot open socket output stream '" +
                           " (" + e.getMessage() + ")");
        }
        this.sock        = sock;
        this.agent       = agent;
        agentName        = agent.getAgentName();
        hostName         = sock.getInetAddress().getHostName();
        port             = sock.getPort();
        connectionIsOpen = true;
    }

    // Methods

    protected void write(String text) throws kqml.KqmlIOException
    {
        try
        {
            out.write(text.getBytes());
            out.flush();
            if (DEBUG)
                Log.log(Log.DEBUG, "write: " + text);
        }
        catch (java.io.IOException e)
        {
            throw new kqml.KqmlIOException(e.getMessage());
        }
    }

    protected String   getLocalHost()
    {
        return (localHost);
    }

    protected Message asyncGet() throws java.io.EOFException,
                                                        java.io.IOException
    {
        Message  inMsg = null;
        Message  rval  = null;

        if ((inIndex < inBuffer.length()) || (in.available() > 0))
        {
            try
            {
                String    buffered = " ";
                Date      start = null;

                inMsg = new Message(this);
                if ((inIndex < inBuffer.length()))
                    buffered = "buffered ";

                if (agent.timingEnabled())
                    start = new Date();

                inMsg.readMessage();

                if (agent.timingEnabled())
                {
                    agent.setKmsgTime(inMsg.getKqmlMsgId(), TimingRecord.READ,
                                                    start, new Date());
                }

                rval = inMsg;         // MUST follow the read
                if (remoteAgent == null)
                {
                    remoteAgent = inMsg.getSender();
                    agent.addAgentToLookup(remoteAgent, this);
                }

		Log.log(Log.READ, "asyncGet: received message;");
		Log.log(Log.READ, inMsg);
		//Log.log(Log.READ, "asyncGet: got " + buffered +
		//  "message from agent " + remoteAgent +
		//  "\n     perf `"   + inMsg.getPerformative() +
		//  "' kqml-msg-id `" + inMsg.getKqmlMsgId() + "'");
            }
            catch(InvalidKQMLTokenException e)
            {
                if (inMsg != null)
                   inMsg.sendResponse(Reply.error("invalid token parse error"));
                Log.log(Log.ERROR, "INVALID TOKEN: " + e.getMessage());
            }
            catch(InvalidKQMLMsgException e)
            {
                if (inMsg != null)
                    inMsg.sendResponse(Reply.error("invalid message error"));
                Log.log(Log.ERROR, "INVALID MSG: " + e.getMessage());
            }
        }

        return (rval);
    }

    private StringBuffer inBuffer = new StringBuffer();
    private int          inIndex = 0;

    protected char  nextInChar() throws InvalidKQMLMsgException
    {
        //
        // Get the next character from the input stream
        //
        if (inIndex >= inBuffer.length())
            readMore();

        if (inIndex < inBuffer.length())
            return (inBuffer.charAt(inIndex++));
        else
            throw new InvalidKQMLMsgException(
                    "At end of kqml input stream for " + remoteAgent);
    }

    protected String nextInChars(int count) throws InvalidKQMLMsgException
    {
        //
        // Get the 'count' next characters from the input stream
        //
        while (inIndex + count > inBuffer.length())
            readMore();

        if (inIndex < inBuffer.length())
        {
            inIndex += count;

            return (inBuffer.toString().substring(inIndex - count, inIndex));
        }
        else
        {
            throw new InvalidKQMLMsgException(
                    "At end of kqml input stream for " + remoteAgent);
        }
    }

    private void readMore()
    {
        //
        // Read the next block of text from the input stream.
        //
        int  firstChar;

        try
        {
            firstChar = in.read();      // wait until a byte is available

            // START Instrumentation
            if (readCalls == 0)
            {
                Log.log(Log.READ+Log.DEBUG, "** readMore: 1st byte [" +
                                    (char) firstChar + "] " + firstChar);
            }
            readCalls++;
            totalBytes = 1;
            // END   Instrumentation
        }
        catch(java.io.IOException e)
        {
            firstChar = -2;
            Log.log(Log.ERROR, "** readMore: IO Exception on stream for " +
                                                                remoteAgent);
        }

        if (firstChar > -1)
        {
            inBuffer.append((char) firstChar);

            try
            {
                int avail;
                while ((avail = in.available()) > 0)
                {
                    int       bytesRead;
                    byte[]    incomingBytes = new byte[avail];

                    bytesRead = in.read(incomingBytes, 0, avail);
                    if (bytesRead > -1)
                    {
                        String  nextTextSection =
                                    new String(incomingBytes, 0, bytesRead);

                        inBuffer.append(nextTextSection);
                        nextTextSection  = null;
                        incomingBytes    = null;

                        // Instrumentation
                        totalBytes      += bytesRead;
                        readCalls++;
                        if (DEBUG)
                            Log.log(Log.MEMORY, "readMore: bytesRead " +
				    bytesRead + " avail " + avail);
                    }
                }
                // Log.log(Log.READ+Log.DEBUG, "** readMore: buffer\n" +
                //                                  inBuffer.toString());
            }
            catch(java.io.IOException e)
            {
                Log.log(Log.ERROR, "** readMore: IO Exception on stream for " +
			remoteAgent);
            }
        }
    }

    protected void  restoreInBuffer(char inChar)
    {
        //
        // Restore the previously retrieved character from the input buffer.
        // NOTE: Assume that the input parameter is the same as the inIndex
        //       character currently in the inBuffer.
        //
        if (inIndex > 0)
            inIndex--;
    }

    //
    // This implementation leaves characters following the closing ")"
    // for subsequent input from "asyncGet"
    //
    protected void finishBuffer()
    {
        int  len = 0;

        if (inIndex < inBuffer.length())
        {
            // Eat a trailing carriage return
            if (inBuffer.charAt(inIndex) == '\r')
                ++inIndex;
        }
        len = inBuffer.length() - startIndex;

        if (inIndex >= inBuffer.length())
            disposeOfInMessage();
        else
            startIndex = inIndex;

        Log.log(Log.READ+Log.DEBUG,
                "** finishBuffer:\n     " + remoteAgent + ":" +
                hostName + ":" + port + " " + len + " bytes in " +
                readCalls + " reads");

        readCalls  = 0;
        totalBytes = 0;
    }

    //
    // This implementation DOES NOT check for trailing by reading until
    // the closing ")".  A more robust implementation would.
    //
    protected void  disposeOfInMessage()
    {
        //
        // Dispose of the message at the beginning of the input buffer.
        //
        inBuffer   = null;
        inBuffer   = new StringBuffer();
        inIndex    = 0;
        startIndex = 0;
    }

    protected void close()
    {
        if (connectionIsOpen)
        {
            try
            {
                in.close();
                Log.log(Log.IO+Log.DEBUG, "** close: close Connection " +
			remoteAgent + ":" + hostName + ":" + port);
            }
            catch(java.io.IOException e)
            {
                Log.log(Log.IO+Log.DEBUG, "** close: close failed for " +
			remoteAgent + ":" + hostName + ":" + port);
                // Take no further action
            }
            closeSock();
            connectionIsOpen = false;
        }
    }

    protected void closeSock()
    {
        if (sock != null)
        {
            try
            {
                sock.close();
                Log.log(Log.IO+Log.DEBUG, "** closeSock: closed socket " +
			remoteAgent + ":" + hostName + ":" + port);
            }
            catch(java.io.IOException e)
            {
                Log.log(Log.IO+Log.DEBUG,
			"** closeSock: closed socket failed for " +
                    remoteAgent + ":" + hostName + ":" + port);
            }
        }
    }

    protected boolean  isOpen()
    {
        return (connectionIsOpen);
    }

    protected String getAgentName()
    {
        return (agentName);
    }

    protected String getHost()
    {
        return (hostName);
    }

    protected KqmlAgent getAgent()
    {
        return (agent);
    }

    protected String getRemoteAgent()
    {
        return (remoteAgent);
    }

    protected void  setRemoteAgent(String remoteagent)
    {
        remoteAgent = remoteagent;
    }

    public String toString()
    {
        String rval = (remoteAgent == null) ? "" : remoteAgent;
        rval += ":" + hostName + ":" + port + "[" + agentName + "]";

        return (rval);
    }
}
