/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         DbInfo.java
**
** Description:  Contains the DbInfoConstants interface and the DbInfo class
**               with the following "public" methods:
**
**                  Constructor to generate keyword value pairs
**                  for KQML messages.  Name and value set and get
**                  methods are also provided.
**
*/
package kqml;
import java.lang.NumberFormatException;
import java.util.Date;

//
// Permits the constants to be used in any class implementing this interface
//
/**
 * An interface defining constants needed to decode a KQML database
 * wrapper SQL select result.
 *
 * @author Peter Stevens
 * @version $Revision: 1.2 $
 * @see dbInfo
 */
interface DbInfoConstants
{
    static final boolean DEBUG          = false;

    public static final int DTYPE_STRING =     1;
    public static final int DTYPE_INT    =     2;
    public static final int DTYPE_FLOAT  =     3;

    static final int MAXCOLUMNS          =   100;
    static final int MAXFLDWIDTH         =  4096;
    static final int ROW_FLD_WIDTH       =     9;
    static final int COL_FLD_WIDTH       =     3;
    static final int COL_TYPE_WIDTH      =     1;
    static final int COL_LEN_WIDTH       =     6;
    static final int COL_SCALE_WIDTH     =     2;
    static final int MAX_SQL_QUERY_LEN   = 32768;
}

/**
 * A representation of the results of a relational database query.
 * It consists of descriptive information (number of rows, number
 * of columns, data type of each column) about the matrix, followed by
 * the actual matrix data which is a C column by R row, matrix (as you
 * would expect).  The base element of a query result is a row, and each
 * row is represented as a Java "vector" of the columns (in that row).
 * Each column (in the row vector) may be one of three Java types:
 *
 *         Long
 *         Double
 *         String
 *
 * The set of types supported can be extended by adding new type
 * accessors, "DTYPE_" constants, defined above and matching type
 * specific "decoders" in the "Decode" method below. This class is 
 * based on the C "kqml" SQL library "decode_sql.c"
 * @see dbInfoConstants
 * @see Row
 */
public class DbInfo implements DbInfoConstants
{
    // Data
    private int[] datalen   = new int[MAXCOLUMNS];  // Used to set the data
    private int[] datascale = new int[MAXCOLUMNS];  // Used to set the data
    private int[] datatype  = new int[MAXCOLUMNS];  // Used to set&read
    private int   num_rows;
    private int   num_cols;
    private Row[] rows;

    // Constructor
    /**
     * Constructor for a "dbInfo" structure, replaces <code>decode_sql_content
     * </code> in the KQML C language KRIL. This constructor decodes the
     * <code>:content</code> field of an SQL query response.  The first
     * part describes the layout of the information, and the following
     * the "layout" is the actual result of the query.
     * 
     * The layout information is as a series of characters with the
     * following format:
     *<code>
     * Field Name   Applies      Starts  Ends   Length
     *             To Column       At     At
     *
     * #rows       entire table     0      8      9
     * #cols       entire table     9     11      3
     *
     * datatype         1          12     12      1
     * datalen          1          13     18      6
     * datascale        1          19     20      2
     *
     * datatype         2          21     21      1
     * datalen          2          22     27      6
     * datascale        2          28     29      2
     *
     * datatype         3          30     30      1
     * datalen          3          31     36      6
     * datascale        3          37     38      2
     * ...
     *</code>
     * etc. until each column, per #cols has been has been "described".
     *
     * The data for the query then follows, with each column in a row
     * occupying "datalen" characters.  For example, if the each row
     * contains three columns, 4, 5, and 6 characters long then a row
     * occupies 15 characters, and if there are ten rows in the query
     * result, the data part contains 150 characters.
     *
     * Once the query result layout information has been "decoded",
     * the data is converted (as necessary) and moved in each column
     * for a row, with the process executed for all the rows output.
     * @param encoding the <code>:content</code> field of a reply to
     * a database wrapper select transaction 
     * @exception InvalidDbInfoException usually a result of
     * an invalid bad select statement, resulting in non-numeric data in
     * the description fields of a db_info result.
     * @see Row
     */
    public DbInfo (String encoding) throws InvalidDbInfoException
    {
        int parse_index;

        //  Parse the row and column counts
        try
        {
            num_rows    = stringToInt(encoding.substring(0, ROW_FLD_WIDTH));
        }
        catch (NumberFormatException e) { logAndThrow("Invalid Row count", e); }

        parse_index = ROW_FLD_WIDTH + COL_FLD_WIDTH;
        
        try
        {
            num_cols = stringToInt(encoding.substring(ROW_FLD_WIDTH,
                                                            parse_index));
        }
        catch (NumberFormatException e) {logAndThrow("Invalid Column count",e);}
        rows        = new Row[num_rows];   // Allocate an array of rows

        //  Parse the column layout information
        for (int i = 0 ; i < num_cols ; ++i)
        {
            try
            {
                datatype[i] = stringToInt(encoding.substring(parse_index,
                                                parse_index + COL_TYPE_WIDTH));
            }
            catch (NumberFormatException e)
            {
                logAndThrow("Invalid datatype field for column " + i, e);
            }
            parse_index += COL_TYPE_WIDTH;
            try
            {
                datalen[i] = stringToInt(encoding.substring(parse_index,
                                                parse_index + COL_LEN_WIDTH));
            }
            catch (NumberFormatException e)
            {
                logAndThrow("Invalid datalength field for column " + i, e);
            }
            parse_index += COL_LEN_WIDTH;
            try
            {
                datascale[i] = stringToInt(encoding.substring(parse_index,
                                parse_index + COL_SCALE_WIDTH));
            }
            catch (NumberFormatException e)
            {
                logAndThrow("Invalid datascale field for column " + i, e);
            }
            parse_index += COL_SCALE_WIDTH;
        }

        //  Parse the data,
        //      for each row
        //          for each column
        for (int i = 0 ; i < num_rows ; ++i)
        {
            Row    r = rows[i] = new Row(num_cols);

            for (int j = 0 ; j < num_cols ; ++j)
            {
                String sub;

                sub = encoding.substring(parse_index, parse_index + datalen[j]);
                switch (datatype[j])
                {
                case DTYPE_INT:
                    try
                    {
                        Long l = Long.valueOf(sub.trim());
                        if (DEBUG)
                            Log.log(Log.DEBUG, "dbInfo Long fld " + j + "(" +
                                            maxStr(sub) + "): "+ l.toString());
                        r.appendLong(l);
                    }
                    catch (NumberFormatException e)
                    {
                        logAndThrow("dbInfo Long fld " + j + ": ", e);
                        r.appendLong(new Long(0));
                    }
                    break;
                case DTYPE_FLOAT:
                    try
                    {
                        Double d = Double.valueOf(sub.trim());
                        Log.log(Log.DEBUG, "dbInfo Dbl fld " + j + ": " +
                                                                d.toString());
                        if (DEBUG)
                            Log.log(Log.DEBUG, "dbInfo Dbl fld " + j + "(" +
                                            maxStr(sub) + "): "+ d.toString());
                        r.appendDouble(d);
                    }
                    catch (NumberFormatException e)
                    {
                        logAndThrow("dbInfo Double fld " + j + ": ", e);
                        r.appendDouble(new Double(0));
                    }
                    break;
                case DTYPE_STRING:
                    if (DEBUG)
                        Log.log(Log.DEBUG, "dbInfo str fld " + j + ": " +
                                                                maxStr(sub));
                    r.appendString(sub);
                    break;
                default:
                    break;
                }
                parse_index += datalen[j];
            }
        }
    }

    // Methods
    /**
     * PrintTuples used to print the dbInfo instance's tuples result from
     * an SQL select.  It is principally used for debugging select statements
     * and mimics the function provided in the C language KRIL function
     * <code>print_tuples</code>.  Results are printed on the standard output.
     */
    public void PrintTuples()
    {
        if (num_rows == 0)
        {
            System.out.println("       <No tuples returned>");
        }
        else
            for (int i = 0 ; i < num_rows ; ++i)
            {
                Row    r = rows[i];
                String s;

                for (int j = 0; j < num_cols ; ++j)
                {
                    //  These outputs need to be "datalen[j]" columns wide
                    switch (datatype[j])
                    {
                    case DTYPE_INT:
                        s = r.getLongAt(j).toString();
                        break;
                    case DTYPE_FLOAT:
                        s = r.getDoubleAt(j).toString();
                        break;
                    case DTYPE_STRING:
                        s = r.getStringAt(j);
                        break;
                    default:
                        s = "";
                        break;
                    }
                    System.out.print(s.trim() + ((j<(num_cols-1)) ? "|" : ""));
                }
                System.out.println("");
            }
    }

    /**
     * getRowCount returns the number of rows in this dbInfo instance.
     * It is equivalent to dbinfo.rows.length
     * @return the number of rows in this dbInfo instance
     */
    public int getRowCount()
    {
        return (num_rows);
    }

    /**
     * getColCount returns the number of columns in this dbInfo instance.
     * @return the number of columns in this dbInfo instance
     */
    public int getColCount()
    {
        return (num_rows);
    }

    /**
     * getColType returns the data type "DTYPE_" value for the specified 
     * @param colNo column number whose data type value is desired.
     * @return specified column's data type.
     */
    public int getColType(int colNo)
    {
        return (datatype[colNo]);
    }

    /**
     * getDate returns the matrix element at the specified row and column
     * as a Date object.
     * @param rowNo matrix element's row number 
     * @param colNo matrix element's column number
     * @exception java.text.ParseException the string does not parse as a
     * valid date
     * @return the specified matrix element as a Date object
     */
    public Date getDate(int rowNo, int colNo) throws java.text.ParseException
    {
        return (rows[rowNo].convertToDateAt(colNo));
    }

    /**
     * getDouble returns the matrix element at the specified row and column
     * as a Double object.
     * @param rowNo matrix element's row number 
     * @param colNo matrix element's column number
     * @return the specified matrix element as a Double object
     */
    public Double getDouble(int rowNo, int colNo)
    {
        return (rows[rowNo].getDoubleAt(colNo));
    }

    /**
     * getLong returns the matrix element at the specified row and column
     * as a Long object.
     * @param rowNo matrix element's row number 
     * @param colNo matrix element's column number
     * @return the specified matrix element as a Long object
     */
    public Long getLong(int rowNo, int colNo)
    {
        return (rows[rowNo].getLongAt(colNo));
    }

    /**
     * getString returns the matrix element at the specified row and column
     * as a String object.
     * @param rowNo matrix element's row number 
     * @param colNo matrix element's column number
     * @return the specified matrix element as a String object
     */
    public String getString(int rowNo, int colNo)
    {
        return (rows[rowNo].getStringAt(colNo));
    }

    private int stringToInt(String s)
    {
        return (Integer.parseInt(s.trim()));
    }

    private String maxStr(String val)
    {
        return  (val.length() > 60 ? val.substring(0, 60) : val);
    }

    private void logAndThrow(String s, Exception e)
                                    throws InvalidDbInfoException
    {
        Log.log(Log.ERROR, s);
        throw new InvalidDbInfoException(s + e.getMessage());
    }
}
