/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         KqmlSql.java
**
** Description:  Contains the KqmlSql class with the following "public"
**               methods:
**
**                 Methods to set the KqmlAgent instance for sending
**                 transactions to and receiving responses from the
**                 database wrapper, send a vector of SQL statements to
**                 the database wrapper, and send a single SQL statement
**                 the database wrapper.
**
*/
package kqml;
import java.util.Vector;
import kqml.KqmlAgent;
import kqml.Message;
import kqml.Log;

/**
 * A class to issue database wrapper transactions at get the resulting
 * response.  The class is loosely modeled after the C KRIL function
 * <code>kqml_send_sql_query</code>.
 *
 * @author Peter Stevens
 * @version $Revision: 1.6 $
 * @see KqmlAgent
 * @see Message
 * @see java.util.Vector
 */
public class KqmlSql
{
    // fields
    private KqmlAgent agent = null;

    // constructor
    /**
     * Creates the KqmlSql instance, recording the agent instance to be
     * used for transmitting database wrapper queries.
     * @param agent the agent instance to be used in sendSql transactions
     * @exception kqml.InvalidKqmlAgentException the agent parameter is null
     * @see KqmlAgent
     */
    public KqmlSql(KqmlAgent agent) throws kqml.InvalidKqmlAgentException
    {
        if (agent == null)
            throw new InvalidKqmlAgentException("agent parameter is null");
        this.agent = agent;
    }

    /**
     * Sends a vector of transactions to the database wrapper agent.
     * If you are using the Oracle database wrapper, use this form to
     * execute a series of <i>insert, delete, or update</i> SQL statements
     * which must be treated at the wrapper as an atomic transactions are
     * guaranteed to be all executed successfully or are rolled back.
     *
     * The message <code>:content</code> result is a series of strings
     * that should be decoded with a KrilWrapper constructor.  Each string
     * will indicate the number of rows affected by the transaction.
     *
     * <b>Note:  this method should not used to execute a series of
     * <i>select</i> statements.</b>
     *
     * @param perf      the performative
     * @param vec       vector of SQL transactions
     * @param destAgent database wrapper's agent name
     * @return the transaction reply message or null if no reply was received
     * @see KrilWrapper
     * @see Message
     * @see java.util.Vector
    */
    public Message sqlSend(String perf, Vector vec, String destAgent)
    {
        Message reply = null;
        try
        {
            Message outMsg = new Message(perf, vec, destAgent, agent);
            Log.log(Log.SQL, "sqlSend(vec): " + perf + " content `" +
                                            outMsg.getContents() + "'");
            reply = outMsg.sendAndAwaitReply();
        }
        catch (KqmlIOException e)         { printException("IO EXCEPTION", e); }
        catch (KQMLMsgTimeout e)          { printException("MSG TIMEOUT",  e); }
        catch (InvalidKQMLMsgException e) { printException("INVALID MSG",  e); }
        catch (InvalidKQMLTokenException e){printException("INVALID TOKEN",e); }

        return (reply);
    }

    /**
     * Sends a single transaction to the database wrapper agent.
     * If you are using the Oracle database wrapper, use this form to
     * send any single SQL data manipulation or query statement including
     * <i>select</i>.  If the transaction is a <i>select</i> then the
     * message<code>:content</code> result should be decoded using a dbInfo
     * constructor, otherwise it will be an ascii string that indicates
     * the number of rows affected.
     *
     * @param perf      the performative
     * @param content   the SQL transaction
     * @param destAgent database wrapper's agent name
     * @return the transaction reply message or null if no reply was received
     * @see Message
     * @see dbInfo
     */
    public Message sqlSend(String perf, String content, String destAgent)
    {
        Message reply = null;

        Log.log(Log.SQL, "sqlSend: send `" + perf + "'" +
                " to `" + destAgent + "'\n  content `" + content + "'");

        try
        {
            reply = agent.sendMsgAndAwaitReply(perf, content, destAgent);
        }
        catch (KqmlIOException e)         { printException("IO EXCEPTION", e); }
        catch (KQMLMsgTimeout e)          { printException("MSG TIMEOUT",  e); }
        catch (InvalidKQMLMsgException e) { printException("INVALID MSG",  e); }
        catch (InvalidKQMLTokenException e){printException("INVALID TOKEN",e); }
        return (reply);
    }
    
    private void printException(String s, Exception e)
    {
        Log.log(Log.ERROR, s + ": " + e.getMessage());
    }
}
