/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         Log.java
**
** Description:  Contains the Log class with the following "public" static
**               methods:
**
**                 setFilename to set name of the log file and
**                 setGeneratedFilename (has the time of day automatically
**                 suffixed to the name), logging level set and get methods,
**                 logging format set and get methods, a method to enter a
**                 message in the log, and finally a method to log a stack
**                 track for an exception.  These replace the corresponding
**                 _kqml_dbg_xxx functions in the C KRIL.
**
*/
package kqml;
import java.util.Date;
import java.util.Hashtable;
import java.util.TimeZone;
import java.text.SimpleDateFormat;
import java.io.PrintWriter;
import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;

/**
 * This class is used to log diagnostic and debugging messages for a
 * KQML agent. Log messages are recorded if the current log level setting
 * matches that in the log level value passed to Log.log.  The Java KRIL
 * uses the Log.log function widely.  The logging levels are encoded
 * bit-wise so that diagnostic messages for different "levels" can be
 * selected individually (that is, they can be set in parallel).
 * <p>
 * The log level may be also be set from the <code>DBG_LEVEL</code>
 * system property at run time, if it is a valid integer (can be decimal,
 * hex[prefixed by 0x or #], or octal [prefixed by a 0]).
 * <p>
 * Log output format is set using "setFormat".  The developer can set
 * the message prefix as follows:
 * <ol>
 *   <li>time of day, to the millisecond
 *   <li>date (yyJJJ) and time of day (JJJ is the julian date)
 *   <li>time since the last message
 *   <li>the message level as an alpha code 
 * </ol><p>
 * The default format setting is "time of day" message prefix.
 * <p>
 * The alpha code setting (<code>LOG_FORMAT_PRINT_CODE</code>) can
 * used to record all levels in the log and then use an after-the-
 * fact filter for codes of interest.  The code names are:
 * <pre>
 *      ERROR   "ERR "
 *      READ    "RD  "
 *      WRITE   "WRT "
 *      SQL     "SQL "
 *      TRACE   "TRC "
 *      MEMORY  "MEM "
 *      DEBUG   "DBG "
 *      THREAD  "THR "
 *      256     "B08 "
 *      512     "B09 "
 *      etc.
 * </pre>
 * <p>
 * Debugging levels are defined as:
 *
 *    <pre>
 *    ERROR  = 1               * error messages
 *    READ   = 2               * read event
 *    WRITE  = 4               * write event
 *    IO     = READ + WRITE
 *    SQL    = 8               * SQL I/O
 *    TRACE  = 16              * not used at present
 *    MEMORY = 32              * log memory use
 *    DEBUG  = 64              * debug API operation (that is, all stages
 *                             * of read and write events are logged)
 *    THREAD = 128             * debug API "thread" dispatching events
 *    ALWAYS = 0xFFFFFFFF      * all log bits set.</pre>
 *
 * <p><b>Programming Notes</b>
 * <p>
 * A call to Log.log, typically, will only set one of the level bits.
 * For example:<pre>
 *
 *    Log.log(Log.ERROR, "an error occurred");
 * </pre>
 * with<pre>
 *
 *    Log.setLevel(Log.ERROR);
 * </pre>
 * will record messages marked with the <code>ERROR</code> bit (including the
 * message "an error occurred" and any from the Java KQML API's) in the log
 * file.  If you were to write:<pre>
 *
 *    Log.log(Log.ERROR+Log.DEBUG, "a trivial error occurred");
 * </pre>
 * then:<pre>
 *
 *    Log.setLevel(Log.ERROR);
 * </pre>
 * would set the logging level to <code>ERROR</code>, but the message
 * "a trivial error occurred" would not appear in the log, unless the
 * logging level were set thus:<pre>
 *
 *    Log.setLevel(Log.ERROR+Log.DEBUG);
 * </pre><p>
 * The log level could designate <b>two</b> levels of code instrumentation
 * for an event.  For example, the output concerning a <code>READ</code> event:
 * <pre>
 *    Log.log(Log.READ, "a read event");
 *    Log.log(Log.READ+Log.DEBUG,"a detailed output concerning the read event");
 * </pre><p>
 * The first message would appear if the logging level were set to READ
 * and the second message would only appear if the logging level were set
 * READ and DEBUG, thus:<pre>
 *
 *    Log.setLevel(Log.READ);           // one message would appear in the log
 *    Log.setLevel(Log.READ+Log.DEBUG); // two messages would appear in the log
 * </pre><p>
 * If you wish to instrument your code to record "user" level events
 * in the log, you should use the bits above 128 (there are 24 bits
 * unassigned).  For example:<pre>
 *
 *    public static int USER_EVENT0 = 256;
 *    public static int USER_EVENT1 = 512;
 * </pre>
 * and the calls to log:<pre>
 *
 *    Log.log(USER_EVENT0, "a user event 0 occurred");
 *    Log.log(USER_EVENT1, "a user event 1 occurred");
 * </pre>
 * with the following call to setLevel:<pre>
 *
 *    Log.setLevel(USER_LEVENT0);
 * </pre>
 * the message "<code>a user event 0 occurred</code>" will appear in the log.
 * <p>
 * With this call to setLevel:<pre>
 *
 *    Log.setLevel(USER_LEVENT1);
 * </pre>
 * the message "<code>a user event 1 occurred</code>" will appear in the log.
 * </pre><p>
 * With this call to setLevel:<pre>
 *
 *    Log.setLevel(USER_LEVENT0 + USER_LEVENT1);
 * </pre>
 * the messages "<code>a user event 0 occurred</code>" and
 * "<code>a user event 1 occurred</code>" will both appear in the log.
 *
 * @author Peter Stevens
 * @author Christian Polizzi
 * @version $Revision: 1.8 $
 * @see Message#toString
 * @see KqmlAgent#toString
 */

public class Log 
{
    // adjust the static initializer
    /**
     * Constants for logging level setting
     */
    /**
     * The value for no logging.
     */
    public static int NONE   = 0;
    /**
     * The value for error logging.
     */
    public static int ERROR  = 1;
    /**
     * The value for read event logging.
     */
    public static int READ   = 2;
    /**
     * The value for write event logging.
     */
    public static int WRITE  = 4;
    /**
     * The value for read and write event logging.
     */
    public static int IO     = READ + WRITE;
    /**
     * The value for SQL event logging.
     */
    public static int SQL    = 8;
    /**
     * The value for "trace" event logging.  Currently not used in the API.
     */
    public static int TRACE  = 16;
    /**
     * The value for memory use logging.
     */
    public static int MEMORY = 32;
    /**
     * The value for debug (i.e. verbose) logging.
     */
    public static int DEBUG  = 64;
    /**
     * The value for thread event logging.
     */
    public static int THREAD = 128;
    /**
     * The value for logging all events.
     */
    public static int ALWAYS = 0xFFFFFFFF;

    private static int MAX_LEVEL = 0x80000000;

    //
    // Note: the settings are bit-wise encoded
    //
    /**
     * The value prints logs with no timestamp
     */
    public  static int LOG_FORMAT_NO_TIMESTAMP        = 0;
    /**
     * The value prints logs with a time of day timestamp
     */
    public  static int LOG_FORMAT_TIME_ONLY           = 1;
    /**
     * The value prints logs with a date and time of day timestamp
     */
    public  static int LOG_FORMAT_FULL_DATE           = 2;
    /**
     * The value prints logs with a delta since the last log message.
     */
    private static int LOG_FORMAT_DELTA               = 4;
    /**
     * The value prints logs with a time of day timestamp
     * and delta since the last log message.
     */
    public  static int LOG_FORMAT_TIME_ONLY_AND_DELTA = 5;
    /**
     * The value prints logs with a date and time of day timestamp
     * and delta since the last log message.
     */
    public  static int LOG_FORMAT_FULL_DATE_AND_DELTA = 6;
    /**
     * The value prints logs with a date and time of day timestamp
     * with a message level "CODE" after the timestamp and before
     * the log message.
     */
    public  static int LOG_FORMAT_PRINT_LEVEL         = 8;

    public static final String DBG_LEVEL        = "DBG_LEVEL";
    public static final String ERROR_LEVEL_NAME = "ERROR";

    private static final Runtime rt = Runtime.getRuntime();

    private static Hashtable   levelName = new Hashtable();
    private static int         logFormat = LOG_FORMAT_TIME_ONLY;
    private static int         logLevel  = ERROR;
    private static PrintWriter out       = null;
    private static Date        prev      = null;
    private static TimeZone    tz        = TimeZone.getDefault();

    private static SimpleDateFormat deltaform;
    private static SimpleDateFormat longDate;
    private static SimpleDateFormat timeOnly;

    static
    {
        String dbgLevel = null;

        longDate  = new SimpleDateFormat("yD.HH:mm:ss.SSS");
        timeOnly  = new SimpleDateFormat("HH:mm:ss.SSS");
        deltaform = new SimpleDateFormat("m:ss.SSS");
        longDate.setTimeZone(tz);
        timeOnly.setTimeZone(tz);
        logLevel  = Integer.getInteger(DBG_LEVEL,new Integer(ERROR)).intValue();

        //
        // Logging level name associative lookup table initialization.
        // These must be adjusted as the level list above is adjusted
        //
        levelName.put(new Integer(ERROR),  "ERR ");
        levelName.put(new Integer(READ),   "RD  ");
        levelName.put(new Integer(WRITE),  "WRT ");
        levelName.put(new Integer(SQL),    "SQL ");
        levelName.put(new Integer(TRACE),  "TRC ");
        levelName.put(new Integer(MEMORY), "MEM ");
        levelName.put(new Integer(DEBUG),  "DBG ");
        levelName.put(new Integer(THREAD), "THR ");

        // These are placeholders - replace them if the levels are expanded
        levelName.put(new Integer(1<<8),   "B08 ");
        levelName.put(new Integer(1<<9),   "B09 ");
        levelName.put(new Integer(1<<10),  "B10 ");
        levelName.put(new Integer(1<<11),  "B11 ");
        levelName.put(new Integer(1<<12),  "B12 ");
        levelName.put(new Integer(1<<13),  "B13 ");
        levelName.put(new Integer(1<<14),  "B14 ");
        levelName.put(new Integer(1<<15),  "B15 ");

        levelName.put(new Integer(1<<16),  "B16 ");
        levelName.put(new Integer(1<<17),  "B17 ");
        levelName.put(new Integer(1<<18),  "B18 ");
        levelName.put(new Integer(1<<19),  "B19 ");
        levelName.put(new Integer(1<<20),  "B20 ");
        levelName.put(new Integer(1<<21),  "B21 ");
        levelName.put(new Integer(1<<22),  "B22 ");
        levelName.put(new Integer(1<<23),  "B23 ");

        levelName.put(new Integer(1<<24),  "B24 ");
        levelName.put(new Integer(1<<25),  "B25 ");
        levelName.put(new Integer(1<<26),  "B26 ");
        levelName.put(new Integer(1<<27),  "B27 ");
        levelName.put(new Integer(1<<28),  "B28 ");
        levelName.put(new Integer(1<<29),  "B29 ");
        levelName.put(new Integer(1<<30),  "B30 ");
        levelName.put(new Integer(1<<31),  "B31 ");
    }

    /**
     * Method to set the logging file.  If this or setGeneratedFilename are
     * never called, the log output is to the standard output.
     * @param filename the file to open for log output.
     * @see #log
     */
    public static void setFilename(String filename) throws IOException
    {
        if (filename != null)
            out = new PrintWriter(new BufferedWriter(new FileWriter(filename)),
                                                                    true);
    }

    /**
     * Method to set the logging file, with the time of suffixed to the name.
     * If this or setFilename are never called, the log output is to the
     * standard output.
     * @param filename the file to open for log output.
     * @see #log
     */
    public static void setGeneratedFilename(String filename) throws IOException
    {
        String fn = genFilename(filename);

        if (fn != null)
            out = new PrintWriter(new BufferedWriter(new FileWriter(fn)), true);
    }

    /**
     * Method to set the logging level.
     * @param level the new logging level value
     */
    public static void setLevel(int level)
    {
        logLevel = level;
    }

    /**
     * Method to get the logging level.
     * @return the current logging level
     */
    public static int getLevel()
    {
        return (logLevel);
    }

    /**
     * Method to set the format control setting.
     * @param format the new format control setting
     */
    public static void setFormat(int format)
    {
        logFormat = format;
    }

    /**
     * Method to get the format control setting.
     * @return the current format control setting
     */
    public static int getFormat()
    {
        return (logFormat);
    }


    /**
     * Method to print a log message if current logging level matches
     * the input logging level (in all the level bits).  A time stamp,
     * potentially containing the year, julian day, the time of day to
     * the milliesecond and a time delta since the last message recorded
     * may be output prior to the message text itself.  If setFilename
     * hasn't been called the log output will be to the standard output.
     *
     * @param level the logging level value for this log output string
     * @param s the log output string
     * @see #setFilename
     * @see #setLevel
     */
    public synchronized static void log(int level, String s)
    {
        if ((logLevel & level) == level)
        {
            Date   currentTime = new Date();
            String entry;

            entry = formatPrefix(level, currentTime) + s;

            if ((level & MEMORY) == MEMORY)
                entry += (" Total mem: " + rt.totalMemory() +" Free mem: " +
                                    rt.freeMemory());

            if (out != null)
                out.println(entry);
            else
                System.out.println(entry);

            prev = currentTime;       // Don't record "prev" time unless the
                                      // message was printed in the log.
        }
    }

    /**
     * Overloading of the log method to print a log message, if
     * the input logging level matches the current logging level (in
     * all the level bits).
     * @param level the logging level value for this log output string
     * @param obj the object whose "toString" method will be invoked for
     * output to the log.  If the object is an exception, the getMessage
     * method will be called to print on the log.
     * @see #log(int, String)
     * @see #setFilename
     * @see #setLevel
     */
    public synchronized static void log(int level, Object obj)
    {
        if ((logLevel & level) == level)
        {
            String s;

            if (obj instanceof Message)
                s = "\n" + ((Message)obj).toString();
            else if (obj instanceof KqmlAgent)
                s = ((KqmlAgent)obj).toString();
            else if (obj instanceof Throwable)
                s = ((Throwable)obj).getMessage();
            else
                s = obj.toString();

            log(level, s);
        }
    }

    /**
     * Method to print an exception stack trace to the log if current
     * logging level matches the input logging level (in at least one bit).
     * A time stamp containing the year, julian day, the time of day to the
     * milliesecond and a time delta since the last message recorded are
     * optionally output prior to the stack trace.  If setFilename hasn't
     * been called the log output will be to the standard error.
     *
     * @param level the logging level value for this log output string
     * @param e the log output string
     * @see #setFilename
     * @see #setFormat
     * @see #setLevel
     */
    public synchronized static void logStackTrace(int level, Throwable e)
    {
        if ((logLevel & level) != 0)
        {
            Date   currentTime = new Date();
            String entry;

            entry = formatPrefix(level, currentTime);

            if (out != null)
            {
                e.printStackTrace(out);
                out.println(entry);
            }
            else
            {
                System.out.println(entry);
                e.printStackTrace();
            }

            prev = currentTime;       // Don't record "prev" time unless the
                                      // message was printed in the log.
        }
    }

    //
    // Generates out log message prefix according to log format control
    //
    private static String formatPrefix(int level, Date currentTime)
    {
        String rval;

        if ((logFormat&LOG_FORMAT_TIME_ONLY) == LOG_FORMAT_TIME_ONLY)
            rval = timeOnly.format(currentTime) + " ";
        else if ((logFormat&LOG_FORMAT_FULL_DATE) == LOG_FORMAT_FULL_DATE)
            rval = longDate.format(currentTime) + " ";
        else
            rval = "";

        if ((logFormat&LOG_FORMAT_DELTA) == LOG_FORMAT_DELTA)
        {
            if (prev == null)
                rval += " START ";
            else
            {
                Date   elapsed;

                elapsed = new Date(currentTime.getTime() - prev.getTime());
                rval  += deltaform.format(elapsed) + " ";
            }
        }
        if ((logFormat&LOG_FORMAT_PRINT_LEVEL) == LOG_FORMAT_PRINT_LEVEL)
        {
            for (int logLevel=1; logLevel<=MAX_LEVEL; logLevel <<= 1)
                if ((logLevel&level) != 0)
                {
                    rval += (String)levelName.get(new Integer(logLevel));
                    break;  // record one level value, even if more are present
                }
        }
        return (rval);
    }

    private static String genFilename(String filename)
    {
        SimpleDateFormat  filenameForm = new SimpleDateFormat("HHmmss");
        String            rval         = filename;

        if (rval != null)
        {
            filenameForm.setTimeZone(tz);
            rval += "." + filenameForm.format(new Date());
        }

        return (rval);
    }
}
