/*
** Restricted Rights Legend
**
** Use, duplication or disclosure by the government is subject to restrictions
** as set forth in subparagraph (c)(1)(ii) of the Rights in Technical Data and
** Computer Software clause at DFARS 252.227-7013.
**
** Copyright (c) 1997 Lockheed Martin C2 Integration Systems
**
**
** File:         Row.java
**
** Description:  Contains the Row class with the following methods:
**
**                 Selectors for String, Long and Double data types,
**                 and a convenience "converter" function for Date objects.
*/
package kqml;
import java.text.DateFormat;
import java.util.Date;
import java.util.Vector;

//
//  A Row is a vector of values; use the Dbinfo class "datatype[j]"
//  to choose the right accessor type (i.e., Long, Double, or String).
//

/**
 * A class to used to represent the columns in a Row returned from 
 * an SQL select statement.  Since the data type of results from
 * a select statement result may be heterogeneous, a Vector is a type
 * neutral mechanism storing the results (is an array of Object faster?).
 *
 * @author Peter Stevens
 * @version $Revision: 1.8 $
 * @see KqmlSql
 * @see java.text.DateFormat;
 * @see java.util.Date
 * @see java.util.Vector
 */
class Row extends Vector
{
    static DateFormat df;

    static
    {
        df = DateFormat.getDateInstance();
    }

    /**
     * 
     * Returns a row object with the specified number of columns.
     *
     * @param number number of columns in the row.
     * @see KqmlSql
     */
    protected Row(int number)
    {
        new Vector(number);
    }

    //
    // Long vector accessors
    //
    /**
     * Returns a Long object from the specified column in the Row
     * instance.  The object at the specified column must be a Long
     * or a ClassCastException will result.  An invalid column number
     * will trigger a NoSuchElementException.
     *
     * @param index the column number (0 based) of the Long to be retrieved
     * @return the Long object at the specified column
     * @see java.lang.ClassCastException
     * @see java.util.Vector#elementAt(int)
     * @see java.util.NoSuchElementException
     */
    protected Long getLongAt(int index)
    {
        return ((Long) elementAt(index));
    }
    /**
     * Enters a Long object into the Row instance at the specified column.
     *
     * @param val   the Long object to be set in the Row instance
     * @param index the column number (0 based) of the column Long to be
     *              retrieved
     * @see java.util.Vector#setElementAt(Object, int)
     */
    protected void setLongAt(Long val, int index)
    {
        setElementAt(val, index);
    }
    /**
     * Inserts a Long object into the Row instance at the specified column.
     * Each component in the Row vector with an index greater or equal to
     * the specified <code>index</code> is shifted upward to have an index
     * one greater than the value it had previously.
     *
     * @param val   the Long object to be set in the Row instance
     * @param index the column number of the column Long to be inserted
     * @see java.util.Vector#insertElementAt(Object, int)
     */
    protected void insertLongAt(Long val, int index)
    {
        insertElementAt(val, index);
    }
    /**
     * Appends a Long object into the Row instance following the last column.
     *
     * @param val the Long object to be set in the Row instance
     * @see java.util.Vector#addElement(Object)
     */
    protected void appendLong(Long val)
    {
        addElement(val);
    }
    
    //
    // Double vector accessors
    //
    /**
     * Returns a Double object from the specified column in the Row
     * instance.  The object at the specified column must be a Double
     * or a ClassCastException will result.  An invalid column number
     * will trigger a NoSuchElementException.
     *
     * @param index the column number (0 based) of the Double to be retrieved
     * @return the Double object at the specified column
     * @see java.lang.ClassCastException
     * @see java.util.Vector#elementAt(int)
     * @see java.util.NoSuchElementException
     */
    protected Double getDoubleAt(int index)
    {
        return ((Double) elementAt(index));
    }
    /**
     * Enters a Double object into the Row instance at the specified column.
     *
     * @param val   the Double object to be set in the Row instance
     * @param index the column number (0 based) of the column Double to be
     *              retrieved
     * @see java.util.Vector#setElementAt(Object, int)
     */
    protected void setDoubleAt(Double val, int index)
    {
        setElementAt(val,index);
    }
    /**
     * Inserts a Double object into the Row instance at the specified column.
     * Each component in the Row vector with an index greater or equal to
     * the specified <code>index</code> is shifted upward to have an index
     * one greater than the value it had previously.
     *
     * @param val   the Double object to be set in the Row instance
     * @param index the column number of the column Double to be inserted
     * @see java.util.Vector#insertElementAt(Object, int)
     */
    protected void insertDoubleAt(Double val, int index)
    {
        insertElementAt(val,index);
    }
    /**
     * Appends a Double object into the Row instance following the last column.
     *
     * @param val the Double object to be set in the Row instance
     * @see java.util.Vector#addElement(Object)
     */
    protected void appendDouble(Double val)
    {
        addElement(val);
    }
    
    //
    // String vector accessors
    //
    /**
     * Returns a String object from the specified column in the Row
     * instance.  The object at the specified column must be a String
     * or a ClassCastException will result.  An invalid column number
     * will trigger a NoSuchElementException.
     *
     * @param index the column number (0 based) of the String to be retrieved
     * @return the String object at the specified column
     * @see java.lang.ClassCastException
     * @see java.util.Vector#elementAt(int)
     * @see java.util.NoSuchElementException
     */
    protected String getStringAt(int index)
    {
        return ((String) elementAt(index));
    }
    /**
     * Enters a String object into the Row instance at the specified column.
     *
     * @param val   the String object to be set in the Row instance
     * @param index the column number (0 based) of the column String to be
     *              retrieved
     * @see java.util.Vector#setElementAt(Object, int)
     */
    protected void setStringAt(String val, int index)
    {
        setElementAt(val,index);
    }
    /**
     * Inserts a String object into the Row instance at the specified column.
     * Each component in the Row vector with an index greater or equal to
     * the specified <code>index</code> is shifted upward to have an index
     * one greater than the value it had previously.
     *
     * @param val   the String object to be set in the Row instance
     * @param index the column number of the column String to be inserted
     * @see java.util.Vector#insertElementAt(Object, int)
     */
    protected void insertStringAt(String val, int index) {
        insertElementAt(val, index);
    }

    /**
     * Appends a String object into the Row instance following the last column.
     *
     * @param val the String object to be set in the Row instance
     * @see java.util.Vector#addElement(Object)
     */
    protected void appendString(String val)
    {
        addElement(val);
    }

    /**
     * Returns a Date object from the specified column in the Row
     * instance.  The object at the specified column must be a String
     * or a ClassCastException will result.  An invalid column number
     * will trigger a NoSuchElementException, and if the string does not
     * form a valid date, a ParseException will be thrown.
     *
     * @param index the column number (0 based) of the Date to be retrieved
     * @return the Date object at the specified column
     * @exception java.text.ParseException if the column string does not form
     * a valid date.
     * @see java.lang.ClassCastException
     * @see java.util.Vector#elementAt(int)
     * @see java.util.NoSuchElementException
     */
    Date convertToDateAt(int index) throws java.text.ParseException
    {
        return (df.parse((String)elementAt(index)));
    }
}
