/* File: KqmlLogReader.java
 * Contains: Class for looking at Kqml log files
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: February 1998
 * Updated: Thu Mar  9 18:56:14 2000 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.acp3;

import java.io.*;

import oplan.util.*;
import oplan.util.lisp.*;


/**
 * A KqmlLogReader can be used to read messages (in list form) from a
 * KQML log file.
 */

public abstract class KqmlLogReader implements KqmlSymbols {

    protected LispReader lin;	// our source
    protected int count = 0;	// the number of messages seen

    /**
     * nextMessage() returns the next message (as a LList), or Lisp.EOF
     * if the end of the log file has been reached.
     */
    public abstract Object nextMessage();

    /**
     * A factory method that returns an instance of an appropriate
     * subclass of KqmlLogReader.
     */
    public static KqmlLogReader newInstance(String filename) 
	             throws FileNotFoundException {
        File f = new File(filename);
	if (f.getName().startsWith("acp3"))
	    return new Acp3LogReader(filename);
	else
	    return new MpaLogReader(filename);
    }

    /**
     * A utility for getting the :delta-t property of a message.
     */
    public static long messageDelta_t(LList message) {
	// /\/: It should just be a field in KqmlEvent.
	LList plist = message.cdr();
	Long delta_t = (Long)plist.get(K_DELTA_T);
	if (delta_t == null)
	    return 0;
	else
	    return delta_t.longValue();
    }

}

class MpaLogReader extends KqmlLogReader {

    protected Object c;		// a piece of file content

    public MpaLogReader(String filename) throws FileNotFoundException {
	Debug.noteln("New MpaLogReader");
	this.lin = new LispFileReader(filename);
    }

    public Object nextMessage() {
	Debug.assert(c != Lisp.EOF, "read past end of Kqml log");
	c = lin.safeRead();
	while (c != Lisp.EOF) {
	    // Look for: sending to <agent name> <message>
	    if (c != S_SENDING) {
		c = lin.safeRead();
		continue;
	    }
	    if (!((c = lin.safeRead()) == S_TO)) continue;
	    if (!((c = lin.safeRead()) instanceof Symbol)) continue;
	    // We have the agent name
	    Symbol sending_to = (Symbol)c;
	    // Now for the message
	    if (!((c = lin.safeRead()) instanceof Cons)) continue;
	    // Got it!
	    count++;
	    // Return the message with a :sent-to property added:
	    Cons m = (Cons)c;
	    return new Cons(m.car(), 			  // the performative
			    new Cons(K_SENT_TO,	          // :sent-to
				     new Cons(sending_to, // agent-name
					      m.cdr()))); // other properties
	}
	Debug.noteln(count + " messages processed.");
	return Lisp.EOF;
    }

}

class Acp3LogReader extends KqmlLogReader {

    public Acp3LogReader(String filename) throws FileNotFoundException {
	Debug.noteln("New Acp3LogReader");
	this.lin = new LispFileReader(filename);

	Object header = lin.safeRead();
	Debug.noteln("Header:", header);

	Debug.assert(header instanceof String, "the header is a String");
    }

    public Object nextMessage() {
	Object message = lin.safeRead();
	Debug.assert(message instanceof Cons || message == Lisp.EOF,
		     "valid Acp3 log contents");
	return message;
    }

}
