/* File: KqmlWatcher.java
 * Contains: A watcher for significant Kqml messages
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Thu Mar  9 18:55:10 2000 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.acp3;

import java.text.SimpleDateFormat;
import java.util.*;
import java.io.*;

import oplan.util.*;
import oplan.util.match.*;
import oplan.util.lisp.*;

import kqml.*;


/**
 * A KqmlWatcher is used to look for significant Kqml messages and
 * tell an implementation of WatcherListener about them.<p>
 *
 * Using a KqmlWatcher typically involves four steps:
 *
 * <pre>
 *    KqmlWatcher watcher = new KqmlWatcher();
 *    watcher.setListener(someWatcherListener);
 *    watcher.init();
 *    watcher.watcherMainLoop();
 * </pre>
 *
 * The listener can also be provided as an argument to new KqmlWatcher.<p>
 *
 * init() creates a Kqml agent called "ACP3" and creates message handlers
 * for the Kqml performatives ACP3 expects to receive.  It also creates a
 * log file.  The call to watcher.init() should be omitted in cases when
 * it is not intended for ACP3 to receive Kqml messages.<p>
 *
 * Assignment to KqmlWatcher.kqmlAgentName before init() is called
 * will change the name used when registering the Kqml agent.<p>
 *
 * watcherMainLoop() loops forever waiting for messages and interpreting
 * them when they arrive.  It should be called only if init() was called
 * earlier.
 *
 * @see WatcherListener
 * @see MessageWatcher
 */

public class KqmlWatcher {

    public    WatcherListener listener;
    public    KqmlAgent kqmlAgent;
    protected KqmlInterpreter interp;
    protected MessageQueue kqmlEvents = new MessageQueue();

    protected Date creationDate = new Date();

    // Public log file into in case something wants to look at it.
    public String logDirectoryName = "acp3-logs";
    public String logFileName;		     // does not include the directory
    public Writer logWriter;

    /**
     * Our name as a Kqml agent
     */
    protected static String kqmlAgentName = "ACP3";

    /**
     * Create a KqmlWatcher connected to a WatcherListener.
     */
    public KqmlWatcher(WatcherListener listener) {
	this.listener = listener;
	this.interp = new Acp3KqmlInterpreter(this);
    }

    /**
     * Change the WatcherListener that the KqmlWatcher tells of status changes.
     */
    public void setListener(WatcherListener listener) {
	this.listener = listener;
	((Acp3KqmlInterpreter)this.interp).setListener(listener);
    }

    /**
     * The init() method creates a KqmlAgent and then registers message
     * handlers for the Kqml performatives we care about.  It also creates
     * a log file.  init() should not be called if there's no desire to
     * receive actual KQML messages.  A KqmlWatcher may still be useful
     * in such cases, because there are other ways to give it messages
     * to interpret.
     */
    public void init() {

	// Debug.assert(listener instanceof WatcherListener, "no listener");

	try {
	    createLogFile();

	    Debug.noteln("Creating Kqml agent");

	    // Create the Kqml agent.
	    kqmlAgent = new KqmlAgent(kqmlAgentName);

	    // Register handler(s).
	    registerHandler("evaluate", true);
	    registerHandler("tell", true);
	    registerHandler("delete", false);
	    registerHandler("insert", false);
	    registerHandler("reply", false);
	    registerHandler("error", false);
	}
        catch (KqmlAgentRegistrationFailure e) { Debug.noteException(e); }
        catch (KqmlAgentInitFailure e)         { Debug.noteException(e); }
	catch (IOException e)                  { Debug.noteException(e); }
    }

    protected void registerHandler(String performative, boolean analyse) {
	kqmlAgent.registerHandler
	    (performative,
	     analyse ? new MessageWatcher(this)
	             : new ReplyHandler(this));
    }

    /**
     * Analyses messages enqueued by pleaseAnalyse.<p>
     *
     * Loops forever.
     */
    public void watcherMainLoop() {
	Debug.assert(listener instanceof WatcherListener, "no listener");
	Debug.assert(kqmlAgent instanceof KqmlAgent, "no Kqml agent");

	Debug.noteln("Starting watcherMainLoop");
	for (;;) {
	    Message msg = (Message)kqmlEvents.nextMessage();
	    analyseMessage(msg);
	    Thread.yield();
	}
    }

    /**
     * Enqueues a Kqml Message for analysis in watcherMainLoop.
     */
    public void pleaseAnalyse(Message msg) {
	kqmlEvents.send(msg);	// here send means enqueue
    }

    /**
     * Performs the actual message analysis.
     */
    protected void analyseMessage(Message msg) {
	KqmlEvent event = new KqmlEvent(msg);
	Debug.noteln("Looking at", event);
	interp.interpret(event);
    }

    /**
     * A public interface to the Kqml message-interpretation code.
     */
    public void interpret(KqmlEvent event) {
	interp.interpret(event);
    }

    /**
     * Create a file for recording the messages we receive.
     * The file name is based on the date and time when createLogFile
     * was called and will be something like "acp3-log-18-mar-98-21:51:05".
     */
    protected void createLogFile() throws IOException {
	SimpleDateFormat sdf = new SimpleDateFormat("dd-MMM-yy-HH:mm:ss");
	Date d;
	File f;
	// We sometimes have to explicitly set the default timezone;
	// See oplan/test/DateTest.java.  /\/
	sdf.setTimeZone(TimeZone.getDefault());
	// Find an unused file name.
	while (true) {
	    d = new Date();	// current date and time
	    logFileName = "acp3-log-" + sdf.format(d);
	    f = new File(logDirectoryName, logFileName);
	    if (!f.exists())
		break;
	    else {
		// This should never happen, but just in case ...
		Debug.warn("Log file " + logFileName + " already exists");
		try { Thread.sleep(1000); }
		catch (InterruptedException e) {}
	    }
	}
	// Create the file
	try {
	    // logWriter = new BufferedWriter(new FileWriter(f));
	    logWriter = new FileWriter(f);
	}
	catch (IOException e) {
	    Debug.noteln("Can't make log file " + logFileName +
			 " in " + logDirectoryName);
	    // fatalException(e);
	    throw (IOException)e.fillInStackTrace();
	}
	// Write a header line as a quoted string.
	try {
	    logWriter.write("\"ACP3 log, watcher started " +
			    sdf.format(creationDate) + "\"\n\n");
	}
	catch (IOException e) {
	    Debug.noteException(e);
	}
    }

    /**
     * Record the message in our log file.
     */
    protected void logMessage(Message msg) {
	Date whenReceived = new Date();
	// Debug.noteln("Received Kqml:", msg);

	long delta_t = whenReceived.getTime() - creationDate.getTime();
	String performative = msg.getPerformative();
	String sender = msg.getSender();
	String reply_with = msg.getReplyWith();
	String content = msg.getContents();
	String toLog
	    = "(" + performative
	          + " :delta-t " + Long.toString(delta_t)
	          + " :sender \"" + sender + "\""
	          + " :reply-with " + reply_with + "\n  "
	          + " :content " + content
	          + ")\n\n";

	// Debug.noteln("Received", performative);

	synchronized (logWriter) {
	    try {
		// Here's where we actually log something.
		logWriter.write(toLog);
		logWriter.flush(); 	// N.B. even when not buffered /\/
	    }
	    catch (IOException e) {
		Debug.noteException(e);
	    }
	}
    }

    /**
     * shutdown should be called before the system exits so that it can
     * call kqmlAgent.shutdown().  The main program class should define a
     * shutdown method that calls our shutdown method, and ACP3.exit will
     * then cause the right things to happen.  That is, we should have
     * something like: ACP3.exit --> main program class shutdown -->
     * KqmlWatcher shutdown --> KqmlAgent shutdown.
     */
    public void shutdown() {
	Debug.noteln("KqmlWatcher shutdown");
	if (logWriter != null) {
	    try { logWriter.close(); }
	    catch (IOException e) {
		Debug.noteException(e);
	    }
	}
	if (kqmlAgent != null) {
	    Debug.noteln("Calling kqml agent shutdown()");
	    kqmlAgent.shutdown();
	}
	else
	    Debug.noteln("No Kqml agent to shut down.");
    }

    /**
     * fatalException is called when an exception indicates that something
     * that must happen has failed, and there isn't anything in the system
     * that would enable it to fix the problem and try again.
     */
    static void fatalException(Exception e) {
	Debug.noteln("\nFatal exception\n");
        Debug.noteException(e);
	ACP3.exit(1);		// will arrange shutdown
    }

}


/*
 * Message handler classes
 */

// /\/: The Kqml code creates a number of different threads that
// receive messages.  When messages arrive very quickly (especially
// when no reply needs to be sent?), it's possible for messages to
// be logged, or given to the watcher via pleaseAnalyse(), in the
// wrong order.  At present, there's synchronisation only for adding
// the KqmlEvent to the watcher.kqmlEvents MessageQueue and for writing
// to the log file, and these are separate synchronisations.  One
// synchronisation around the everything the handler does would minimise
// the problem -- at least it would ensure that the log file and the
// watcher's message queue had things in the same order -- but probably
// not completely eliminate it.  Not wanting to make such a change so late
// in the day, I've instead turned off Debug output in the message
// logging code (so it will be faster) and changed the MessageWatcher
// handler() method to call watcher.pleaseAnalyse(msg) before logging,
// rather than after.

// Note that this problem as so far not been observed in actual use,
// only in testing using SendKqml; and the changes described above
// have, so far, been sufficient.

/**
 * A Kqml Message handler that just sends an :OK reply to messages
 * that expect a reply.  Note that the MPA spec requires that an :OK
 * or :ERROR reply be sent.
 */
class ReplyHandler implements Handler {

    protected KqmlWatcher watcher;

    /**
     * Create a ReplyHandler for a particular KqmlWatcher.
     */
    ReplyHandler(KqmlWatcher watcher) {
	this.watcher = watcher;
    }

    /**
     * Handles a Kqml message by returning a standard reply,
     * if a reply is required.
     *
     * @see ReplyHandler#makeReply
     */
    public Reply handler(Message msg) {
	watcher.logMessage(msg);
	return makeReply(msg);
    }

    /**
     * makeReply is responsible for making a "standard" reply to a message.
     * It assumes that if we receive a message with :content of the
     * form (<plan-performative> ...), then (a) the :reply-with property
     * will be <plan-performative>, and (b) we can reply with Kqml
     * performative "reply" and content (<plan-performative> :ok).
     * Assumption (a) allows us to get the <plan-performative> from
     * the :reply-with property rather than from the :content.
     */
    static protected Reply makeReply(Message msg) {
	// Is a reply called for?
	// The :reply-with field is automatically send back as :in-reply-to.
	// It's not clear what we'll ever get :reply-with t or nil.
	String reply_with = msg.getReplyWith();
	// Debug.noteln("Making reply to", msg);
	if (reply_with == null)
	    return null;
	else if (reply_with.equalsIgnoreCase("nil"))
	    return null;
	else if (reply_with.equalsIgnoreCase("t"))
	    return Reply.reply(":ok");
	else
	    return Reply.reply("(" + reply_with + " :ok)");
    }

}

/**
 * A Kqml Message handler that, in addition to sending any required reply,
 * asks the KqmlWatcher to analyse the message.
 *
 * @see ReplyHandler
 */
class MessageWatcher extends ReplyHandler {

    /**
     * Create a MessageWatcher for a particular KqmlWatcher.
     */
    MessageWatcher(KqmlWatcher watcher) {
	super(watcher);
    }

    /**
     * Handles a Kqml message by enqueuing it for analysis by the
     * KqmlWatcher's watcherMainLoop, then returning a standard reply,
     * if a reply is required.
     *
     * @see MessageWatcher#makeReply
     */
    public Reply handler(Message msg) {
	watcher.pleaseAnalyse(msg);
	// could instead do watcher.analyseMessage(msg);
	// but then we should log it first.
	watcher.logMessage(msg);
	return makeReply(msg);
    }

}



/*
 * Interpreting Kqml
 */


/**
 * A KqmlInterpreter is used to match against KqmlEvents.<p>
 *
 * The match(Object) method should not be called.  Instead, call
 * interpret(KqmlEvent).<p>
 *
 * If a MatchCase is added, it must be an instance of KqmlCase (or one
 * of its subclasses).  The pattern in a KqmlCase is a LList that begins
 * with a Keyword, and the KqmlInterpreter indexes the cases by these
 * pattern keywords for more efficient matching.
 *
 * @see KqmlEvent
 * @see KqmlCase
 */

class KqmlInterpreter extends MatchTable {

    MultiHashtable contentIndex = new MultiHashtable();

    KqmlInterpreter() { }

    public synchronized void interpret(KqmlEvent event) {
	// Debug.noteln("Interpreting", event);
	try { match(event); }
	catch (Exception ex) {
	    Debug.noteln("Exception while interpreting", event);
	    Debug.noteException(ex);
	}
    }

    public void addCase(MatchCase c) {
	Debug.assert(c instanceof KqmlCase,
		     "KqmlInterpreter cases must be KqmlCases");
	super.addCase(c);

	Cons p = (Cons)c.pattern;
	Keyword k = (Keyword)p.car();
	contentIndex.put(k, c);
    }

    public Object match(Object data) {
	// Assumes there's no default case. /\/
	KqmlEvent event = (KqmlEvent)data;
	Vector possible = (Vector)contentIndex.get(event.content.car());
	return possible == null ? null : matchCases(data, possible);
    }

}


/**
 * KqmlCase provides a tryMatch method for KqmlEvents.  The KqmlCase pattern
 * must be a LList that begins with a Keyword.
 *
 * @see KqmlEvent
 * @see KqmlInterpreter
 * @see oplan.util.lisp.LList
 * @see oplan.util.lisp.Keyword
 */

class KqmlCase extends MatchCase {

    /**
     * The KqmlEvent sender's name must begin with this string,
     * e.g. "pcm" to allow the sender to be "pcm-1", "pcm-2", etc.
     * This can be "" if all senders are allowed.
     */
    String senderType;

    /**
     * The kqml performative required.  At present, this doesn't tell
     * us very much, because a given plan performative is always sent
     * with the same Kqml performative.
     */
    String kqmlPerformative;

    /**
     * Create a case for a KqmlInterpreter.  This is normally done by
     * creating an anonymous local class that defines an ifSelected
     * method.
     */
    KqmlCase(String senderType, String kqmlPerformative, String pattern) {
	this.senderType = senderType;
	this.kqmlPerformative = kqmlPerformative;
	this.pattern = Lisp.readFromString(pattern);
	Debug.assert(this.pattern instanceof Cons);
	Debug.assert(((Cons)this.pattern).car() instanceof Keyword);
    }

    /**
     * See if the KqmlEvent matches this case by checking the sender
     * and content fields.  If there is a match, fill in the event's
     * bindings field and return the event; otherwise, return null.
     */
    public Object tryMatch(Object data) {
	KqmlEvent event = (KqmlEvent)data;
	// Check sender.
 	if (!event.sender.startsWith(senderType)) {
	    // Debug.noteln("Sender \"" + event.sender +
	    //              "\" does not match \"" + senderType + "\"");
 	    return null;
	}
	// Check content.
	MatchEnv env = SimpleMatcher.match(pattern, event.content);
	if (env != null) {
	    // Have a match
	    event.bindings = env;
	    // Check Kqml performative
	    if (!event.performative.equals(kqmlPerformative)) {
		Debug.warn("Kqml perf " + event.performative +
			   " should be " + kqmlPerformative +
			   " for " + event.content);
		return null;
	    }
	    return event;
	}
	else 
	    return null;
    }

}

/**
 * The KqmlInterpreter for ACP3.
 */

class Acp3KqmlInterpreter extends KqmlInterpreter implements KqmlSymbols {

    protected WatcherListener listener;

    protected Hashtable pcmCoaAssignments = new Hashtable();

    Acp3KqmlInterpreter(KqmlWatcher watcher) {
	this.listener = watcher.listener;
	// Debug.noteln("listener", listener);
    }

    void setListener(WatcherListener listener) {
	this.listener = listener;
    }

    // Instance initializer
    {

	// This comes *after* the initial :init-problem, which is a
	// but late for how we're interpreting it.  /\/
// 	addCase(new KqmlCase(
// 	    "pcm", "tell",
// 	    "(:advice ?advice :mode ?kwd)") {

// 	    public Object ifSelected(Object data, Object matchResult) {
// 		KqmlEvent event = (KqmlEvent)data;
// 		String pcm = event.sender;
// 		int coa = pcmCoa(pcm);
// 		listener.finishAction("formalize planning advice", coa);
// 		return event;
// 	    }
// 	});

	// The initial :init-problem
	addCase(new KqmlCase(
	    "pcm", "evaluate",
	    "(:init-problem :task ?task :plan ?plan)") {

	    public Object ifSelected(Object data, Object matchResult) {
		KqmlEvent event = (KqmlEvent)data;
		String pcm = event.sender;
		String plan = (String)event.bindings.get(Q_PLAN);
		int coa = planCoa(plan);
		warnIf(plan.endsWith("-phase3"),
		       ":init-problem for -phase3 without action net");
		setPcmCoa(pcm, coa);
		listener.startAction("create COA", coa);
		return event;
	    }
	});

	// The :init-problem for the support plan
	// Older: the message can be (... :phantoms :no), hence the &rest.
	addCase(new KqmlCase(
	    "pcm", "evaluate",
	    "(:init-problem :task ?task :plan ?plan" +
	    " :action-network ?an &rest ?properties)") {

	    public Object ifSelected(Object data, Object matchResult) {
		KqmlEvent event = (KqmlEvent)data;
		String pcm = event.sender;
		String plan = (String)event.bindings.get(Q_PLAN);
		int coa = planCoa(plan);
		warnIf(!plan.endsWith("-phase3"),
		       ":init-problem with action net but not -phase3");
		expectPcmCoa(pcm, coa);
		listener.startAction("create support plan", coa);
		return event;
	    }
	});

	// After an :init-problem, there are a number of :expand-plan
	// messages for increasing levels.  This case recognizes the
	// :expand-plans for both the COA and support plans.
	addCase(new KqmlCase(
	    "pcm", "evaluate",
	    "(:expand-plan :task ?task :plan ?plan :action-network ?an)") {
	    
	    public Object ifSelected(Object data, Object matchResult) {
		KqmlEvent event = (KqmlEvent)data;
		String pcm = event.sender;
		String plan = (String)event.bindings.get(Q_PLAN);
		int coa = planCoa(plan);
		String an = event.bindings.get(Q_AN).toString();
		int level = actionNetLevel(an);
		String action;
		if (plan.endsWith("-phase3")) {
		    action = "create support plan to one more level";
		    // action = "execute support-plan-cycle";
		    // post-ctem isn't there in older demos /\/
		    // warnIf(an.indexOf("-post-ctem-") == -1,
		    //        ":expand support plan but not post-ctem");
		}
		else {
		    action = "create COA to one more level";
		    // action = "execute plan-cycle";
		    // post-ctem isn't there in older demos /\/
		    // warnIf(an.indexOf("-post-ctem-") != -1,
		    //        ":expand post-ctem but not support-plan");
		}
		expectPcmCoa(pcm, coa);
		listener.startAction(action, coa, level);
		return event;
	    }
	});

	// After expansion to the max level, we see (:plan-ok? ... :final t).
	// This case handles both the COA and support plans.
	addCase(new KqmlCase(
	    "pcm", "evaluate",
            "(:plan-ok? :task ?task :plan ?plan :action-network ?an" +
	    " :final t)") {

	    public Object ifSelected(Object data, Object matchResult) {
		KqmlEvent event = (KqmlEvent)data;
		String pcm = event.sender;
		String plan = (String)event.bindings.get(Q_PLAN);
		int coa = planCoa(plan);
		String action;
		// We won't check for "-post-ctem-" in ?an this time.  /\/
		if (plan.endsWith("-phase3"))
		    action = "approve final support plan";
		else
		    action = "approve final COA";
		expectPcmCoa(pcm, coa);
		listener.startAction(action, coa);
		return event;
	    }
	});

	// :schedule-ctem comes right between the (:plan-ok? ... :final t)
	// for the COA plan and the :init-problem for the support plan.
	addCase(new KqmlCase(
	    "pcm", "evaluate",
            "(:schedule-ctem :task ?task :plan ?plan :action-network ?an)") {

	    public Object ifSelected(Object data, Object matchResult) {
		KqmlEvent event = (KqmlEvent)data;
		String pcm = event.sender;
		String plan = (String)event.bindings.get(Q_PLAN);
		int coa = planCoa(plan);
		expectPcmCoa(pcm, coa);
		listener.startAction("assign weapons systems", coa);
		return event;
	    }
	});

	// :solution is the last thing we see for a given COA.
	// It's a pcm sending to meta-pcm.
	// In older demos, it was (... :plan-server ?server :id ?id).
	addCase(new KqmlCase(
	    "pcm", "tell",
            "(:solution :task ?task :plan ?plan &rest ?properties)") {

	    public Object ifSelected(Object data, Object matchResult) {
		KqmlEvent event = (KqmlEvent)data;
		String pcm = event.sender;
		String plan = (String)event.bindings.get(Q_PLAN);
		int coa = planCoa(plan);
		expectPcmCoa(pcm, coa);
		listener.finishAction("develop COA", coa);
		return event;
	    }
	});

    }

    protected int pcmCoa(String pcm) {
	Object value = pcmCoaAssignments.get(pcm);
	// return value == null ? new Integer(0) : (Integer)value;
	if (value == null) {
	    Debug.warn("No coa for pcm: " + pcm);
	    return 0;
	}
	else {
	    return ((Integer)value).intValue();
	}
    }

    protected void setPcmCoa(String pcm, int coa) {
	pcmCoaAssignments.put(pcm, new Integer(coa));
    }

    protected void expectPcmCoa(String pcm, int coa) {
	Object value = pcmCoaAssignments.get(pcm);
	if (value == null)
	    Debug.warn("No " + pcm + " coa when expecting " + coa);
	else if (((Integer)value).intValue() != coa)
	    Debug.warn("Found " + pcm + " coa = " + value
		       + " when expecting " + coa);
    }

    protected int planCoa(String name) {
	// E.g. "plan-advice3" ==> 3, "plan-advice2-phase3" ==> 2.
	// Older demos had "plan-2-phase1" ==> 2,
	// "plan-2-phase3" ==> 3, "PLAN-phase1" or "PLAN-phase3 ==> 1.
        Vector parts = Seq.toVector(new StringTokenizer(name, "-"));
	String part0 = (String)parts.elementAt(0);
	if (part0.equalsIgnoreCase("plan")) {
	    String part1 = (String)parts.elementAt(1);
	    String digits;
	    int coa;
	    if (part1.startsWith("advice"))
		// e.g. "plan-advice2-phase3", digits = "2"
		digits = part1.substring("advice".length());
	    else if (part1.startsWith("phase"))
		// e.g. "PLAN-phase3"
		return 1;
	    else if (part1.length() > 0 && Character.isDigit(part1.charAt(0)))
		// e.g. "plan-2-phase3", digits = "2"
		digits = part1;
	    else {
		Debug.warn("Can't find coa in " + name);
		return -1;
	    }
	    try { coa = Integer.parseInt(digits); }
	    catch (NumberFormatException e) {
		Debug.warn("Bad coa int in plan advice name: " + name);
		coa = -1;
	    }
	    return coa;
	}
    Debug.warn("Strange plan-advice name: " + name);
    return -1;
    }

    protected int actionNetLevel(String an) {
	// E.g. as-a-v1-4 or as-a-v1-15-post-ctem-v2-4 ==> 4.
	// In older demos: as1-v1-4 or as1-v1-8-v2-4.
	Vector parts = Seq.toVector(new StringTokenizer(an, "-"));
	return safeInt((String)parts.lastElement(), -1);
    }

    protected static void warnIf(boolean condition, String message) {
	// This method exists because we may not want to do Debug.warn.
	if (condition) Debug.warn(message);
    }

    protected static int safeInt(String digits, int errval) {
	int i;
	try { i = Integer.parseInt(digits); }
	catch (NumberFormatException e) {
	    Debug.warn("Bad int in \"" + digits + "\"");
	    i = errval;
	}
	return i;
    }

}

