/* File: Util.java
 * Contains: A class for useful static methods
 * Author: Jeff Dalton <J.Dalton@ed.ac.uk>
 * Created: January 1998
 * Updated: Sun Apr 12 01:15:16 1998 by Jeff Dalton
 * Copyright: (c) 1998, AIAI, University of Edinburgh
 */

package oplan.util;

import java.util.*;
import java.io.*;


/** Class for useful static methods. */

public class Util {

    /**
     * breakStringAtFirst takes a string containing fields separated by
     * a (string) delimiter and returns a two-element string array containing
     * the substring before the first occurrence of the char, and the
     * substring after.  Neither substring contains the delimiter.  If
     * the delimiter does not appear in the string at all, the values
     * are the string and "".
     */
    public static String[] breakStringAtFirst(String s, String separator) {
	int i = s.indexOf(separator);
	if (i == -1)
	    return new String[]{s, ""};
	else
	    return new String[]{
		s.substring(0, i),
	        s.substring(i + separator.length())
	    };
    }


    /**
     * Print the elements of a String[] array to System.out as lines.
     */
    public static void printLines(String[] lines) {
	for (int i = 0; i < lines.length; i++) {
	    System.out.println(lines[i]);
	}
    }


    /**
     * Simple, text-based user interaction.  askLine pronts a prompt
     * to System.out and returns a String containing the next line
     * from System.in.<p>
     *
     * If askLine blocks when reading, we'd like other threads to be
     * able to run; but that doesn't seem to happen reliably.  Presumably,
     * this is a bug.  In any case, askLine works around the problem by
     * having a loop that checks whether input is available and sleeps
     * for a second if it isn't.
     */
    public static String askLine(String prompt) {
	System.out.print(prompt + " ");
	System.out.flush();
	// /\/: Should we synchronize on System.in?
	try {
	    while (System.in.available() == 0) {
		try { Thread.sleep(1000); }
		catch (InterruptedException e) {}
	    }
	}
	catch (IOException e) {
	    Debug.noteException(e);
	    return "";
	}
	return Util.readLine(System.in);
    }


    /** 
     * Reads a line from an InputStream and returns it as a String.
     * In Java, we seem to have to write this ourself unless we wrap
     * a special stream (or Reader) around whatever we want to read.
     * Here we provide a static method, because that's easier to
     * mix with other operations.  The only InputStream method
     * called is read().
     */
    public static String readLine(InputStream is) {
	ByteArrayOutputStream line = new ByteArrayOutputStream(80);

    readLoop:
	while (true) {
	    int c;		// a character as a "byte"
	    try { c = is.read(); }
	    catch (IOException e) {
		Debug.warn("IOException reading line from " + is);
		break readLoop;
	    }
	    if (c == '\n') 	// newline
		break readLoop;
	    else if (c == -1)	// end of stream
		break readLoop;
	    else if (c == '\r') // CR
		;		// (ignore)
	    else
		line.write(c);
	}
	return line.toString();
    }


}
